package dev.fitko.fitconnect.api.services;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;
import dev.fitko.fitconnect.api.exceptions.AttachmentUploadException;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;

import java.util.UUID;

/**
 * A technical system that creates a submission via the FIT-Connect Submission API.
 * <p>
 * The Sender acts as a common interface wrapping all client functionality for authenticating <p>
 * and creating a valid {@link SubmitSubmission} including encrypted {@link Data} and {@link Attachment}s
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/sending/overview">Sending Submissions</a>
 */
public interface Sender {

    /**
     * Validates the public key consisting of the following steps:
     * <p>
     * <ul>
     *     <li>checks if the JSON Web Key is suitable for the encryption</li>
     *     <li>checks if the public key is matching the certificate referenced in the JWK</li>
     *     <li>checks the certificate chain up to the root certificate</li>
     *     <li>checks against a certificate revocation list and/or an OSCP-endpoint with signed response</li>
     * </ul>
     * @param publicKey the public JWK
     *
     * @return {@link ValidationResult} that includes an error if the validation failed
     */
    ValidationResult validatePublicKey(RSAKey publicKey);

    /**
     * Validates the {@link Metadata} structure against a given JSON-schema to ensure its correctness.
     *
     * @param metadata the {@link Metadata} object that is validated
     * @return a {@link ValidationResult}, contains an error if the {@link Metadata} is invalid or doesn't match the schema
     */
    ValidationResult validateMetadata(Metadata metadata, String schema);

    /**
     * Encrypts the submission data payload (json or xml) with JWE (JSON-Web-Encryption).
     *
     * @param publicKey the public encryption key for encryption of {@link Data}
     * @param data      the {@link Data} payload that should be encrypted
     * @return the JWE-encrypted {@link Data}
     *
     * @throws EncryptionException if the encryption fails due to an invalid key or corrupt data
     */
    String encryptBytes(RSAKey publicKey, byte[] data) throws EncryptionException;


    /**
     * Encrypts the submission data payload (json or xml) with JWE (JSON-Web-Encryption).
     *
     * @param publicKey the public encryption key
     * @param metadata  the data payload that should be encrypted
     * @return the JWE-encrypted {@link Data}
     *
     * @throws EncryptionException if the encryption fails due to an invalid key or corrupt data
     */
    String encryptObject(RSAKey publicKey, Object metadata) throws EncryptionException;

    /**
     * Create message digest (SHA-512) of a given byte[]
     * @param data input bytes that should be hashed
     * @return string representation of hashed byte[]
     */
    String createHash(byte[] data);


    /**
     * Creates and announces a new {@link CreateSubmission}.
     *
     * @param submission with a destinationId, a list of attachmentIds and a serviceType
     * @return the announced submission
     *
     * @throws SubmissionNotCreatedException if the announcement of the submission failed, e.g. due to a technical problem
     */
    SubmissionForPickup createSubmission(CreateSubmission submission) throws SubmissionNotCreatedException;

    /**
     * Uploads encrypted {@link Attachment}s to for given submission id.
     *
     * @param submissionId unique identifier of the  submission
     * @param attachmentId unique identifier of the attachment
     * @param encryptedAttachment string of JWE encrypted attachment
     *
     * @throws AttachmentUploadException
     */
    void uploadAttachment(UUID submissionId, UUID attachmentId, String encryptedAttachment);

    /**
     * Retrieves the public encryption key for a destination.
     *
     * @param destinationId unique identifier of the destination
     * @param encryptionKid key id of the encryption key
     * @return the public key for encrypting {@link Metadata}, {@link Data} and {@link Attachment}s
     *
     * @throws KeyNotRetrievedException if a technical problem occurred fetching the key
     */
    RSAKey getEncryptionKeyForDestination(UUID destinationId, String encryptionKid) throws KeyNotRetrievedException;

    /**
     * Send a {@link SubmitSubmission} that includes encrypted {@link Metadata} and {@link Data}
     * @param submission submission to submit
     */
    void sendSubmission(SubmitSubmission submission);


    /**
     * Gets {@link Destination} by id.
     *
     * @param destinationId unique identifier of a {@link Destination}
     * @return the destination
     */
    Destination getDestination(UUID destinationId);

    /**
     * Sends a rejection event if the submission violates any validation rule.
     *
     * @param submissionId unique identifier of submission
     * @param destinationId unique identifier of destination
     * @param caseId unique identifier of case
     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void rejectSubmission(UUID submissionId, UUID destinationId, UUID caseId);
}
