package dev.fitko.fitconnect.core;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.jwk.JWK;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.MetadataNotCreatedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.events.SecurityEventTokenService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.Collections;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

class SubmissionSenderTest {

    private DestinationService destinationServiceMock;
    private SubmissionService submissionServiceMock;
    private EventLogService eventLogServiceMock;
    private CryptoService cryptoServiceMock;
    private ValidationService validationServiceMock;
    private SecurityEventTokenService setEventServiceMock;

    private Sender underTest;

    @BeforeEach
    public void setUp() {
        this.destinationServiceMock = mock(DestinationService.class);
        this.submissionServiceMock = mock(SubmissionService.class);
        this.eventLogServiceMock = mock(EventLogService.class);
        this.cryptoServiceMock = mock(CryptoService.class);
        this.validationServiceMock = mock(ValidationService.class);
        this.setEventServiceMock = mock(SecurityEventTokenService.class);

        underTest = new SubmissionSender(
                destinationServiceMock,
                submissionServiceMock,
                eventLogServiceMock,
                cryptoServiceMock,
                validationServiceMock,
                setEventServiceMock
        );
    }

    @Test
    void validatePublicKey() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(validationServiceMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(publicKey);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void encryptBytes() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(cryptoServiceMock.encryptBytes(any(), any())).thenReturn("encryptedString");

        // When
        final String encryptedData = underTest.encryptBytes(publicKey, "data".getBytes());

        // Then
        assertNotNull(encryptedData);
        assertThat(encryptedData, is("encryptedString"));
    }

    @Test
    void createHash() {

        // Given
        when(cryptoServiceMock.hashBytes(any())).thenReturn("hashedData");

        // When
        final String hashedData = underTest.createHash("data".getBytes());

        // Then
        assertNotNull(hashedData);
        assertThat(hashedData, is("hashedData"));
    }

    @Test
    void createSubmission() {

        // Given
        final var createSubmission = CreateSubmission.builder()
                .caseId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .announcedAttachments(Collections.emptyList())
                .serviceType(new ServiceType())
                .build();

        final var expectedSubmission = new SubmissionForPickup();
        expectedSubmission.setSubmissionId(UUID.randomUUID());
        expectedSubmission.setDestinationId(createSubmission.getDestinationId());
        expectedSubmission.setCaseId(createSubmission.getCaseId());

        when(submissionServiceMock.announceSubmission(any())).thenReturn(expectedSubmission);

        // When
        final SubmissionForPickup submissionForPickup = underTest.createSubmission(createSubmission);

        // Then
        assertThat(submissionForPickup, equalTo(expectedSubmission));
    }

    @Test
    void uploadAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "encryptedAttachmentBytes";

        doNothing().when(submissionServiceMock).uploadAttachment(any(), any(), any());

        // When
        underTest.uploadAttachment(submissionId, attachmentId, encryptedAttachment);

        // Then
        verify(submissionServiceMock, times(1)).uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Test
    void getEncryptionKeyForDestination() throws ParseException, IOException {

        // Given
        final var submissionId = UUID.randomUUID();

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
        final JWK jwk = new ObjectMapper().readValue(publicKeyJson, JWK.class);
        final RSAKey expectedRSAKey = RSAKey.parse(publicKeyJson);

        when(destinationServiceMock.getEncryptionKey(any(), any())).thenReturn(jwk);

        // When
        final RSAKey key = underTest.getEncryptionKeyForDestination(submissionId, jwk.getKid());

        // Then
        assertThat(key, equalTo(expectedRSAKey));
    }

    @Test
    void getEncryptionKeyForDestinationFailed() {

        // Given
        when(destinationServiceMock.getEncryptionKey(any(), any())).thenThrow(RestApiException.class);

        // Then
        assertThrows(
                KeyNotRetrievedException.class,
                () -> underTest.getEncryptionKeyForDestination(UUID.randomUUID(), "123"),
                "Expected getEncryptionKeyForDestination() to throw, but nothing was thrown"
        );
    }


    @Test
    void encryptObject() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(cryptoServiceMock.encryptString(any(), any())).thenReturn("encryptedObject");

        final Metadata metadata = new Metadata();

        // When
        final String encryptedMetadata = underTest.encryptObject(publicKey, metadata);

        // Then
        assertThat(encryptedMetadata, equalTo("encryptedObject"));
    }

    @Test
    void encryptingObjectFailed() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        // When
        when(cryptoServiceMock.encryptString(any(), any())).thenThrow(EncryptionException.class);

        // Then
        assertThrows(
                MetadataNotCreatedException.class,
                () -> underTest.encryptObject(publicKey, new Metadata()),
                "Expected encryptObject() to throw, but nothing was thrown"
        );
    }

    @Test
    void sendSubmission() {

        // Given
        final SubmitSubmission submitSubmission = new SubmitSubmission();
        submitSubmission.setSubmissionId(UUID.randomUUID());
        submitSubmission.setEncryptedData("encryptedData");
        submitSubmission.setEncryptedMetadata("encryptedMetadata");

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submitSubmission.getSubmissionId());
        expectedSubmission.setEncryptedData(submitSubmission.getEncryptedData());
        expectedSubmission.setEncryptedMetadata(submitSubmission.getEncryptedMetadata());

        when(submissionServiceMock.sendSubmission(any())).thenReturn(expectedSubmission);

        // When
        underTest.sendSubmission(submitSubmission);

        // Then
        verify(submissionServiceMock, times(1)).sendSubmission(submitSubmission);
    }

    @Test
    void getDestination() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var expectedDestination = new Destination();
        expectedDestination.setDestinationId(destinationId);

        when(destinationServiceMock.getDestination(any())).thenReturn(expectedDestination);

        // When
        final Destination destination = underTest.getDestination(destinationId);

        // Then
        assertThat(destination, equalTo(expectedDestination));
    }

    @Test
    void testValidMetadata() {

        // Given
        final var metadata = new Metadata();

        when(validationServiceMock.validateMetadataSchema(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validatedMetadata = underTest.validateMetadata(metadata, "metadataSchema");

        // Then
        assertTrue(validatedMetadata.isValid());
    }

    @Test
    void testInvalidMetadata() {

        // Given
        final var invalidMetadata = new Metadata();
        final ValidationResult expectedResult = ValidationResult.error(new ValidationException("invalid metadata"));
        when(validationServiceMock.validateMetadataSchema(any(), any())).thenReturn(expectedResult);

        // When
        final ValidationResult validatedMetadata = underTest.validateMetadata(invalidMetadata, "metadataSchema");

        // Then
        assertTrue(validatedMetadata.hasError());
        assertThat(validatedMetadata.getError().getMessage(), equalTo("invalid metadata"));
    }

    private String getResourceAsString(final String filename) throws IOException {
        return new String(SubmissionSenderTest.class.getResourceAsStream(filename).readAllBytes());
    }
}